@@ -1,32 +1,33 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright 2021 Nitrokey Developers
+# Copyright 2021-2022 Nitrokey Developers
 #
 # Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
 # http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
 # http://opensource.org/licenses/MIT>, at your option. This file may not be
 # copied, modified, or distributed except according to those terms.
 
-import itertools
 import logging
 import os.path
 import platform
 import time
-from concurrent.futures import ThreadPoolExecutor
 from typing import List, Optional, TypeVar
 
 import click
+from spsdk.mboot.exceptions import McuBootConnectionError
 
 from pynitrokey.helpers import ProgressBar, local_critical, local_print
 from pynitrokey.nk3 import list as list_nk3
 from pynitrokey.nk3 import open as open_nk3
 from pynitrokey.nk3.base import Nitrokey3Base
 from pynitrokey.nk3.bootloader import (
+    RKHT,
     FirmwareMetadata,
     Nitrokey3Bootloader,
     check_firmware_image,
 )
 from pynitrokey.nk3.device import BootMode, Nitrokey3Device
+from pynitrokey.nk3.exceptions import TimeoutException
 from pynitrokey.nk3.updates import get_latest_update, get_update
 from pynitrokey.nk3.utils import Version
 
@@ -115,7 +116,13 @@ def reboot(ctx: Context, bootloader: bool) -> None:
                 local_print(
                     "Please press the touch button to reboot the device into bootloader mode ..."
                 )
-                device.reboot(BootMode.BOOTROM)
+                try:
+                    device.reboot(BootMode.BOOTROM)
+                except TimeoutException:
+                    local_critical(
+                        "The reboot was not confirmed with the touch button.",
+                        support_hint=False,
+                    )
             else:
                 local_critical(
                     "A Nitrokey 3 device in bootloader mode can only reboot into firmware mode."
@@ -236,6 +243,32 @@ def fetch_update(path: str, force: bool, version: Optional[str]) -> None:
         local_critical(f"Failed to download firmware update {update.tag}", e)
 
 
+@nk3.command()
+@click.argument("image")
+def validate_update(image: str) -> None:
+    """
+    Validates the given firmware image and prints the firmware version and the signer.
+    """
+    with open(image, "rb") as f:
+        data = f.read()
+
+    try:
+        metadata = FirmwareMetadata.from_image_data(data)
+    except Exception as e:
+        local_critical("Failed to parse and validate firmware image", e)
+
+    if metadata.rkht:
+        if metadata.rkht == RKHT:
+            signature = "Nitrokey"
+        else:
+            signature = f"unknown issuer (RKHT: {metadata.rkht.hex()})"
+    else:
+        signature = "unsigned"
+
+    print(f"version:    {metadata.version}")
+    print(f"signed by:  {signature}")
+
+
 @nk3.command()
 @click.argument("image", required=False)
 @click.option(
@@ -326,7 +359,13 @@ def update(ctx: Context, image: Optional[str], experimental: bool) -> None:
             local_print(
                 "Please press the touch button to reboot the device into bootloader mode ..."
             )
-            device.reboot(BootMode.BOOTROM)
+            try:
+                device.reboot(BootMode.BOOTROM)
+            except TimeoutException:
+                local_critical(
+                    "The reboot was not confirmed with the touch button.",
+                    support_hint=False,
+                )
 
             local_print("")
 
@@ -339,8 +378,29 @@ def update(ctx: Context, image: Optional[str], experimental: bool) -> None:
                 )
                 raise click.Abort()
 
-            with _await_bootloader(ctx) as bootloader:
-                _perform_update(bootloader, data)
+            retries = 3
+            exc = None
+            for i in range(retries):
+                logger.debug(
+                    f"Trying to connect to bootloader, try {i + 1} of {retries}"
+                )
+                try:
+                    with _await_bootloader(ctx) as bootloader:
+                        _perform_update(bootloader, data)
+                    break
+                except McuBootConnectionError as e:
+                    logger.debug("Received connection error", exc_info=True)
+                    exc = e
+                    if i + 1 < retries:
+                        time.sleep(0.5)
+            else:
+                msgs = ["Failed to connect to Nitrokey 3 bootloader"]
+                if platform.system() == "Linux":
+                    msgs += ["Are the Nitrokey udev rules installed and active?"]
+                local_critical(
+                    *msgs,
+                    exc,
+                )
         elif isinstance(device, Nitrokey3Bootloader):
             _print_version_warning(metadata)
             _print_update_warning()
@@ -463,25 +523,18 @@ def _print_update_warning() -> None:
 
 def _perform_update(device: Nitrokey3Bootloader, image: bytes) -> None:
     logger.debug("Starting firmware update")
+    with ProgressBar(
+        desc="Performing firmware update", unit="B", unit_scale=True
+    ) as bar:
+        result = device.update(image, callback=bar.update_sum)
+    logger.debug(f"Firmware update finished with status {device.status}")
 
-    with ThreadPoolExecutor() as executor:
-        indicators = itertools.cycle(["/", "-", "\\", "|"])
-        future = executor.submit(device.update, image)
-        while not future.done():
-            print(
-                f"\r[{next(indicators)}] Performing firmware update "
-                "(may take several minutes) ... ",
-                end="",
-            )
-            time.sleep(0.1)
-        print("done")
-
-        if future.result():
-            logger.debug("Firmware update finished successfully")
-            device.reboot()
-        else:
-            (code, message) = device.status
-            local_critical(f"Firmware update failed with status code {code}: {message}")
+    if result:
+        logger.debug("Firmware update finished successfully")
+        device.reboot()
+    else:
+        (code, message) = device.status
+        local_critical(f"Firmware update failed with status code {code}: {message}")
 
 
 @nk3.command()
