@@ -1,7 +1,6 @@
+import pathlib
 from cx_Freeze import setup, Executable
 
-# Dependencies are automatically detected, but it might need
-# fine tuning.
 buildOptions = dict(packages = [], excludes = [])
 
 msiOptions = dict(
@@ -15,8 +14,11 @@ executables = [
     Executable('nitropy.py', base=base)
 ]
 
+__version_path__ = pathlib.Path(__file__).parent.resolve().absolute() / "pynitrokey" / "VERSION"
+__version__ = open(__version_path__).read().strip()
+
 setup(name='pynitrokey',
-      version = '0.4.0',
+      version = __version__,
       description = 'Nitrokey Python Tools',
       options = dict(build_exe = buildOptions,
                      bdist_msi = msiOptions),
